import numpy as np
import pytest
from numpy.testing import assert_array_equal, assert_array_almost_equal
from scipy.stats import norm

from cal_tools.jungfrau.jungfrau_ff import (
    chunk_multi,
    fill_histogram,
    histogram_errors,
    rebin_histo,
    set_histo_range,
    _peak_position,
)


def test_peak_detection_correctness():
    x = np.array([10, 20, 30, 40, 50, 60, 70, 80])
    h = np.array([0, 2, 0, 4, 0, 6, 0, 8])
    peaks, thr_idx = _peak_position(x, h)
    assert peaks == [7]
    assert thr_idx == 7


def test_non_equal_length_inputs():
    x = np.array([10, 20, 30])
    h = np.array([1, 2])
    with pytest.raises(AttributeError):
        _peak_position(x, h)


def test_threshold_effectiveness():
    x = np.array([10, 20, 30, 40, 50])
    h = np.array([100, 200, 50, 200, 100])
    thr = 25  # Set threshold to exclude the first two peaks
    peaks, thr_idx = _peak_position(x, h, thr=thr)
    assert peaks == [3]
    assert thr_idx == 2


def test_distance_constraint():
    x = np.array([0, 20, 40, 60, 80, 100, 120, 140, 160, 180, 200])
    h = np.array([1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1])
    dist = 20  # Minimum distance between peaks
    peaks, thr_idx = _peak_position(x, h, dist=dist)
    assert peaks == [9, 7, 5]
    assert thr_idx == 4


@pytest.fixture
def sample_chunk():
    return np.random.rand(100, 1, 256, 64).astype(np.float32)  # 100 trains, 1 cell, 256x64 chunk


@pytest.fixture
def sample_chunk_16_cells():
    return np.random.rand(100, 16, 256, 64).astype(np.float32)  # 100 trains, 16 cells, 256x64 chunk


def test_fill_histogram_basic(sample_chunk):
    h_bins = np.linspace(0, 1000, 101, dtype=np.float32)  # 100 bins
    hist, edges = fill_histogram(sample_chunk, h_bins)
    
    assert hist.shape == (100, 1, 256, 64)
    assert edges.shape == (101,)
    assert np.all(hist >= 0)
    assert np.sum(hist) == np.prod(sample_chunk.shape)


def test_fill_histogram_16_cells(sample_chunk_16_cells):
    h_bins = np.linspace(0, 1000, 101, dtype=np.float32)
    hist, _ = fill_histogram(sample_chunk_16_cells, h_bins)
    
    assert hist.shape == (100, 16, 256, 64)
    assert np.sum(hist) == np.prod(sample_chunk_16_cells.shape)


def test_fill_histogram_single_train():
    chunk = np.random.rand(1, 1, 256, 64).astype(np.float32)
    h_bins = np.linspace(0, 100, 11, dtype=np.float32)
    hist, _ = fill_histogram(chunk, h_bins)
    
    assert hist.shape == (10, 1, 256, 64)
    assert np.sum(hist) == np.prod(chunk.shape)


def test_fill_histogram_single_bin():
    chunk = np.ones((10, 1, 256, 64), dtype=np.float32)
    h_bins = np.array([0, 2], dtype=np.float32)
    hist, _ = fill_histogram(chunk, h_bins)
    
    assert hist.shape == (1, 1, 256, 64)
    assert np.all(hist == 10)


def test_fill_histogram_float_data():
    chunk = np.random.rand(50, 1, 256, 64).astype(np.float32)
    h_bins = np.linspace(0, 1, 11, dtype=np.float32)
    hist, _ = fill_histogram(chunk, h_bins)
    
    assert hist.shape == (10, 1, 256, 64)
    assert np.sum(hist) == np.prod(chunk.shape)


def test_fill_histogram_out_of_range():
    chunk = np.random.rand(100, 1, 256, 64).astype(np.float32)
    h_bins = np.linspace(0, 100, 11, dtype=np.float32)
    hist, _ = fill_histogram(chunk, h_bins)
    
    assert hist.shape == (10, 1, 256, 64)
    assert np.sum(hist) <= np.prod(chunk.shape)


@pytest.mark.parametrize("shape", [
    (100, 1, 256, 64),
    (50, 16, 256, 64),
    (200, 1, 128, 32)
])
def test_fill_histogram_various_shapes(shape):
    chunk = np.random.rand(*shape).astype(np.float32)
    h_bins = np.linspace(0, 1000, 101, dtype=np.float32)
    hist, _ = fill_histogram(chunk, h_bins)
    
    assert hist.shape == (100, *shape[1:])
    assert np.sum(hist) == np.prod(shape)


@pytest.fixture
def sample_data():
    # 100 trains, 1 memory cell, 1024x512 image
    return np.random.rand(100, 1, 1024, 512)


@pytest.fixture
def sample_data_16_cells():
    # 100 trains, 16 memory cells, 1024x512 image
    return np.random.rand(100, 16, 1024, 512)


def test_chunk_multi_basic(sample_data):
    block_size = [256, 64]
    chunks = chunk_multi(sample_data, block_size)
    assert len(chunks) == 32  # (1024 // 256) * (512 // 64) = 4 * 8 = 32
    assert chunks[0].shape == (100, 1, 256, 64)


def test_chunk_multi_16_cells(sample_data_16_cells):
    block_size = [256, 64]
    chunks = chunk_multi(sample_data_16_cells, block_size)
    assert len(chunks) == 32
    assert chunks[0].shape == (100, 16, 256, 64)


def test_chunk_multi_uneven():
    data = np.random.rand(50, 1, 1000, 500)  # Uneven dimensions
    block_size = [256, 64]
    chunks = chunk_multi(data, block_size)
    assert chunks[-1].shape == (50, 1, 232, 52)  # Last chunk should be smaller


def test_chunk_multi_single_train():
    data = np.random.rand(1, 1, 1024, 512)
    block_size = [256, 64]
    chunks = chunk_multi(data, block_size)
    assert len(chunks) == 32
    assert chunks[0].shape == (1, 1, 256, 64)


def test_chunk_multi_small_block():
    data = np.random.rand(10, 1, 1024, 512)
    block_size = [128, 32]
    chunks = chunk_multi(data, block_size)
    assert len(chunks) == 128  # (1024 // 128) * (512 // 32) = 8 * 16 = 128
    assert chunks[0].shape == (10, 1, 128, 32)


@pytest.mark.parametrize("invalid_input", [
    (np.random.rand(1024), [256]),  # 1D array
    (np.random.rand(100, 1, 1024, 512), [0, 64]),  # Invalid block size
    (np.random.rand(100, 1, 1024, 512), [256, 0]),  # Invalid block size
    (np.random.rand(100, 1, 1024, 512), [1025, 64]),  # Block size larger than input
    (np.random.rand(100, 1, 1024, 512), [256]),  # Incomplete block size
    ([1, 2, 3, 4], [256, 64]),  # Not a numpy array
])
def test_chunk_multi_invalid_input(invalid_input):
    with pytest.raises(ValueError):
        chunk_multi(*invalid_input)


def test_chunk_multi_empty_array():
    data = np.array([])
    block_size = [256, 64]
    with pytest.raises(ValueError):
        chunk_multi(data, block_size)


def test_chunk_multi_large_array():
    data = np.random.rand(1000, 1, 1024, 512)  # 1000 trains
    block_size = [256, 64]
    chunks = chunk_multi(data, block_size)
    assert len(chunks) == 32
    assert chunks[0].shape == (1000, 1, 256, 64)


def test_histogram_errors():
    # Test case 1: Array with positive, negative, and zero values
    input_array = np.array([4, -9, 0, 16, -25, 1])
    expected_output = np.array([2, 3, 1, 4, 5, 1])
    assert_array_equal(histogram_errors(input_array), expected_output)

    # Test case 2: Array with all positive values
    input_array = np.array([1, 4, 9, 16])
    expected_output = np.array([1, 2, 3, 4])
    assert_array_equal(histogram_errors(input_array), expected_output)

    # Test case 3: Array with all negative values
    input_array = np.array([-1, -4, -9, -16])
    expected_output = np.array([1, 2, 3, 4])
    assert_array_equal(histogram_errors(input_array), expected_output)

    # Test case 4: Array with all zeros
    input_array = np.zeros(5)
    expected_output = np.ones(5)
    assert_array_equal(histogram_errors(input_array), expected_output)

    # Test case 5: Empty array
    input_array = np.array([])
    expected_output = np.array([])
    assert_array_equal(histogram_errors(input_array), expected_output)

    # Test case 6: Large values
    input_array = np.array([1e6, -1e6])
    expected_output = np.array([1e3, 1e3])
    assert_array_almost_equal(histogram_errors(input_array), expected_output)


    # Test case 7: Small values
    input_array = np.array([1e-6, -1e-6])
    expected_output = np.array([1e-3, 1e-3])
    assert_array_almost_equal(histogram_errors(input_array), expected_output)


def test_rebin_histo():
    """
    Test function for rebin_histo.
    """
    # Test case 1: Successful rebinning
    hist = np.array([1, 2, 3, 4, 5, 6])
    bin_edges = np.array([0, 1, 2, 3, 4, 5, 6])
    rebin_factor = 2
    expected_hist = np.array([3, 7, 11])
    expected_bin_edges = np.array([0, 2, 4, 6])
    h_out, x_out = rebin_histo(hist, bin_edges, rebin_factor)
    assert_array_equal(h_out, expected_hist)
    assert_array_equal(x_out, expected_bin_edges)

    # Test case 2: Rebinning with factor 3
    hist = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    bin_edges = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    rebin_factor = 3
    expected_hist = np.array([6, 15, 24])
    expected_bin_edges = np.array([0, 3, 6, 9])
    h_out, x_out = rebin_histo(hist, bin_edges, rebin_factor)
    assert_array_equal(h_out, expected_hist)
    assert_array_equal(x_out, expected_bin_edges)

    # Test case 3: Error case - rebin_factor is not an exact divisor of the number of bin edges
    hist = np.array([1, 2, 3, 4, 5])
    bin_edges = np.array([0, 1, 2, 3, 4, 5])
    rebin_factor = 2
    with pytest.raises(ValueError):
        rebin_histo(hist, bin_edges, rebin_factor)


def test_set_histo_range_gaussian():
    # Generate a Gaussian-like histogram
    x = np.linspace(-5, 5, 1000)
    h = norm.pdf(x, loc=0, scale=1)
    h_range = (-2, 2)

    # Apply set_histo_range
    new_x, new_h = set_histo_range(x, h, h_range)

    # Check if the new range is within the specified range
    assert new_x[0] >= h_range[0] and new_x[-1] <= h_range[1]

    # Check if the peak is preserved (should be near 0 for this Gaussian)
    assert abs(x[np.argmax(h)] - new_x[np.argmax(new_h)]) < 0.1
